@@ -15,6 +15,8 @@ use crate::binary::{long_align, U16Be, U32Be};
 use crate::cff::CFF;
 use crate::error::{ParseError, ReadWriteError, WriteError};
 use crate::post::PostTable;
+use crate::tables::cmap::owned;
+use crate::tables::cmap::subset::{CmapTarget, MappingsToKeep, NewIds, OldIds};
 use crate::tables::glyf::GlyfTable;
 use crate::tables::loca::{self, LocaTable};
 use crate::tables::{
@@ -28,7 +30,10 @@ pub(crate) trait SubsetGlyphs {
     fn len(&self) -> usize;
 
     /// Return the old glyph id for the supplied new glyph id
-    fn old_id(&self, index: usize) -> u16;
+    fn old_id(&self, new_id: u16) -> u16;
+
+    /// Return the new glyph id for the supplied old glyph id
+    fn new_id(&self, old_id: u16) -> u16;
 }
 
 struct FontBuilder {
@@ -53,50 +58,31 @@ struct OrderedTables {
 }
 
 /// Subset this font so that it only contains the glyphs with the supplied `glyph_ids`.
+///
+/// `glyph_ids` requirements:
+///
+/// * Glyph id 0, corresponding to the `.notdef` glyph must always be present.
+/// * There must be no duplicate glyph ids.
+///
+/// If either of these requirements are not upheld this function will return
+/// `ParseError::BadValue`.
 pub fn subset(
     provider: &impl FontTableProvider,
     glyph_ids: &[u16],
-    cmap0: Option<Box<[u8; 256]>>,
 ) -> Result<Vec<u8>, ReadWriteError> {
+    let mappings_to_keep = MappingsToKeep::new(provider, glyph_ids, CmapTarget::Unrestricted)?;
     if provider.has_table(tag::CFF) {
-        subset_cff(provider, glyph_ids, cmap0, true)
+        subset_cff(provider, glyph_ids, mappings_to_keep, true)
     } else {
-        subset_ttf(provider, glyph_ids, cmap0)
-    }
-}
-
-// TODO: put behind prince feature
-/// Subset this font so that it only contains the glyphs with the supplied `glyph_ids`.
-///
-/// Returns just the CFF table in the case of a CFF font, not a complete OpenType font.
-pub fn prince_subset(
-    provider: &impl FontTableProvider,
-    glyph_ids: &[u16],
-    cmap0: Option<Box<[u8; 256]>>,
-    convert_cff_to_cid_if_more_than_255_glyphs: bool,
-) -> Result<Vec<u8>, ReadWriteError> {
-    if provider.has_table(tag::CFF) {
-        subset_cff_table(
-            provider,
-            glyph_ids,
-            cmap0,
-            convert_cff_to_cid_if_more_than_255_glyphs,
-        )
-    } else {
-        subset_ttf(provider, glyph_ids, cmap0)
+        subset_ttf(provider, glyph_ids, mappings_to_keep)
     }
 }
 
 fn subset_ttf(
     provider: &impl FontTableProvider,
     glyph_ids: &[u16],
-    cmap0: Option<Box<[u8; 256]>>,
+    mappings_to_keep: MappingsToKeep<OldIds>,
 ) -> Result<Vec<u8>, ReadWriteError> {
-    if glyph_ids.get(0) != Some(&0) {
-        // glyph index 0 is the .notdef glyph, the fallback, it must always be first
-        return Err(ReadWriteError::Write(WriteError::BadValue));
-    }
-
     let head = ReadScope::new(&provider.read_table_data(tag::HEAD)?).read::<HeadTable>()?;
     let mut maxp = ReadScope::new(&provider.read_table_data(tag::MAXP)?).read::<MaxpTable>()?;
     let loca_data = provider.read_table_data(tag::LOCA)?;
@@ -120,6 +106,10 @@ fn subset_ttf(
 
     // Subset the glyphs
     let subset_glyphs = glyf.subset(glyph_ids)?;
+    let mappings_to_keep = mappings_to_keep.update_to_new_ids(&subset_glyphs);
+
+    // Build a new cmap table
+    let cmap = create_cmap_table(&mappings_to_keep)?;
 
     // Build new maxp table
     let num_glyphs = u16::try_from(subset_glyphs.len()).map_err(ParseError::from)?;
@@ -143,11 +133,7 @@ fn subset_ttf(
 
     // Build the new font
     let mut builder = FontBuilder::new(0x00010000_u32);
-    if let Some(cmap0) = cmap0 {
-        // Build a new cmap table
-        let cmap = create_cmap_table(glyph_ids, cmap0)?;
-        builder.add_table::<_, cmap::owned::Cmap>(tag::CMAP, cmap, ())?;
-    }
+    builder.add_table::<_, cmap::owned::Cmap>(tag::CMAP, cmap, ())?;
     if let Some(cvt) = cvt {
         builder.add_table::<_, ReadScope<'_>>(tag::CVT, ReadScope::new(&cvt), ())?;
     }
@@ -172,7 +158,7 @@ fn subset_ttf(
 fn subset_cff(
     provider: &impl FontTableProvider,
     glyph_ids: &[u16],
-    cmap0: Option<Box<[u8; 256]>>,
+    mappings_to_keep: MappingsToKeep<OldIds>,
     convert_cff_to_cid_if_more_than_255_glyphs: bool,
 ) -> Result<Vec<u8>, ReadWriteError> {
     let cff_data = provider.read_table_data(tag::CFF)?;
@@ -200,9 +186,13 @@ fn subset_cff(
 
     // Build the new CFF table
     let cff_subset = cff.subset(glyph_ids, convert_cff_to_cid_if_more_than_255_glyphs)?;
+    let mappings_to_keep = mappings_to_keep.update_to_new_ids(&cff_subset);
+
+    // Build a new cmap table
+    let cmap = create_cmap_table(&mappings_to_keep)?;
 
     // Build new maxp table
-    let num_glyphs = u16::try_from(cff_subset.len()).map_err(ParseError::from)?; // FIXME: is .len() correct here
+    let num_glyphs = u16::try_from(cff_subset.len()).map_err(ParseError::from)?;
     maxp.num_glyphs = num_glyphs;
 
     // Build new hhea table
@@ -212,9 +202,6 @@ fn subset_cff(
     // Build new hmtx table
     let hmtx = create_hmtx_table(&hmtx, num_h_metrics, &cff_subset)?;
 
-    // Extract the new CFF table now that we're done with cff_subset
-    let cff = CFF::from(cff_subset);
-
     // Get the remaining tables
     let cvt = provider.table_data(tag::CVT)?;
     let fpgm = provider.table_data(tag::FPGM)?;
@@ -224,11 +211,7 @@ fn subset_cff(
 
     // Build the new font
     let mut builder = FontBuilder::new(tag::OTTO);
-    if let Some(cmap0) = cmap0 {
-        // Build a new cmap table
-        let cmap = create_cmap_table(glyph_ids, cmap0)?;
-        builder.add_table::<_, cmap::owned::Cmap>(tag::CMAP, cmap, ())?;
-    }
+    builder.add_table::<_, cmap::owned::Cmap>(tag::CMAP, cmap, ())?;
     if let Some(cvt) = cvt {
         builder.add_table::<_, ReadScope<'_>>(tag::CVT, ReadScope::new(&cvt), ())?;
     }
@@ -246,37 +229,21 @@ fn subset_cff(
     if let Some(prep) = prep {
         builder.add_table::<_, ReadScope<'_>>(tag::PREP, ReadScope::new(&prep), ())?;
     }
+
+    // Extract the new CFF table now that we're done with cff_subset
+    let cff = CFF::from(cff_subset);
     builder.add_table::<_, CFF<'_>>(tag::CFF, &cff, ())?;
     let builder = builder.add_head_table(&head)?;
     builder.data()
 }
 
-/// Subset the CFF table and discard the rest
-///
-/// Useful for PDF because a CFF table can be embedded directly without the need to wrap it in
-/// an OTF.
-fn subset_cff_table(
-    provider: &impl FontTableProvider,
-    glyph_ids: &[u16],
-    _cmap0: Option<Box<[u8; 256]>>,
-    convert_cff_to_cid_if_more_than_255_glyphs: bool,
-) -> Result<Vec<u8>, ReadWriteError> {
-    let cff_data = provider.read_table_data(tag::CFF)?;
-    let scope = ReadScope::new(&cff_data);
-    let cff: CFF<'_> = scope.read::<CFF<'_>>()?;
-    if cff.name_index.count != 1 || cff.fonts.len() != 1 {
-        return Err(ReadWriteError::from(ParseError::BadIndex));
-    }
-
-    // Build the new CFF table
-    let cff = cff
-        .subset(glyph_ids, convert_cff_to_cid_if_more_than_255_glyphs)?
-        .into();
-
-    let mut buffer = WriteBuffer::new();
-    CFF::write(&mut buffer, &cff)?;
-
-    Ok(buffer.into_inner())
+fn create_cmap_table(
+    mappings_to_keep: &MappingsToKeep<NewIds>,
+) -> Result<owned::Cmap, ReadWriteError> {
+    let encoding_record = owned::EncodingRecord::from_mappings(&mappings_to_keep)?;
+    Ok(owned::Cmap {
+        encoding_records: vec![encoding_record],
+    })
 }
 
 /// Construct a complete font from the supplied provider and tags.
@@ -324,28 +291,6 @@ pub fn whole_font<F: FontTableProvider>(
     builder_with_head.data()
 }
 
-fn create_cmap_table(
-    glyph_ids: &[u16],
-    cmap0: Box<[u8; 256]>,
-) -> Result<cmap::owned::Cmap, ReadWriteError> {
-    use cmap::owned::{Cmap, CmapSubtable, EncodingRecord};
-
-    if glyph_ids.len() > 256 {
-        return Err(ReadWriteError::Write(WriteError::BadValue));
-    }
-
-    Ok(Cmap {
-        encoding_records: vec![EncodingRecord {
-            platform_id: 1, // Macintosh platform
-            encoding_id: 0, // Roman
-            sub_table: CmapSubtable::Format0 {
-                language: 0, // the subtable is language independent
-                glyph_id_array: cmap0,
-            },
-        }],
-    })
-}
-
 fn create_hmtx_table<'b>(
     hmtx: &HmtxTable<'_>,
     num_h_metrics: usize,
@@ -354,7 +299,8 @@ fn create_hmtx_table<'b>(
     let mut h_metrics = Vec::with_capacity(num_h_metrics);
 
     for glyph_id in 0..subset_glyphs.len() {
-        let old_id = usize::from(subset_glyphs.old_id(glyph_id));
+        // Cast is safe as glyph indexes are 16-bit values
+        let old_id = usize::from(subset_glyphs.old_id(glyph_id as u16));
 
         if old_id < num_h_metrics {
             h_metrics.push(hmtx.h_metrics.read_item(old_id)?);
@@ -533,10 +479,75 @@ fn max_power_of_2(num: u16) -> u16 {
     15u16.saturating_sub(num.leading_zeros() as u16)
 }
 
+/// Prince specific subsetting behaviour.
+///
+/// prince::subset will produce a bare CFF table in the case of an input CFF font.
+#[cfg(feature = "prince")]
+pub mod prince {
+    use super::{
+        tag, CmapTarget, FontTableProvider, MappingsToKeep, ParseError, ReadScope, ReadWriteError,
+        WriteBinary, WriteBuffer, CFF,
+    };
+
+    /// Subset this font so that it only contains the glyphs with the supplied `glyph_ids`.
+    ///
+    /// Returns just the CFF table in the case of a CFF font, not a complete OpenType font.
+    pub fn subset(
+        provider: &impl FontTableProvider,
+        glyph_ids: &[u16],
+        macroman_cmap: bool,
+        convert_cff_to_cid_if_more_than_255_glyphs: bool,
+    ) -> Result<Vec<u8>, ReadWriteError> {
+        let cmap_target = if macroman_cmap {
+            CmapTarget::MacRoman
+        } else {
+            CmapTarget::Unrestricted
+        };
+        let mappings_to_keep = MappingsToKeep::new(provider, glyph_ids, cmap_target)?;
+        if provider.has_table(tag::CFF) {
+            subset_cff_table(
+                provider,
+                glyph_ids,
+                convert_cff_to_cid_if_more_than_255_glyphs,
+            )
+        } else {
+            super::subset_ttf(provider, glyph_ids, mappings_to_keep)
+        }
+    }
+
+    /// Subset the CFF table and discard the rest
+    ///
+    /// Useful for PDF because a CFF table can be embedded directly without the need to wrap it in
+    /// an OTF.
+    fn subset_cff_table(
+        provider: &impl FontTableProvider,
+        glyph_ids: &[u16],
+        convert_cff_to_cid_if_more_than_255_glyphs: bool,
+    ) -> Result<Vec<u8>, ReadWriteError> {
+        let cff_data = provider.read_table_data(tag::CFF)?;
+        let scope = ReadScope::new(&cff_data);
+        let cff: CFF<'_> = scope.read::<CFF<'_>>()?;
+        if cff.name_index.count != 1 || cff.fonts.len() != 1 {
+            return Err(ReadWriteError::from(ParseError::BadIndex));
+        }
+
+        // Build the new CFF table
+        let cff = cff
+            .subset(glyph_ids, convert_cff_to_cid_if_more_than_255_glyphs)?
+            .into();
+
+        let mut buffer = WriteBuffer::new();
+        CFF::write(&mut buffer, &cff)?;
+
+        Ok(buffer.into_inner())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
     use crate::font_data::FontData;
+    use crate::tables::cmap::CmapSubtable;
     use crate::tables::glyf::GlyphData;
     use crate::tables::glyf::{
         BoundingBox, CompositeGlyph, CompositeGlyphArgument, CompositeGlyphFlag, GlyfRecord, Glyph,
@@ -545,6 +556,7 @@ mod tests {
     use crate::tables::{LongHorMetric, OpenTypeData, OpenTypeFont};
     use crate::tag::DisplayTag;
     use crate::tests::read_fixture;
+    use crate::Font;
 
     use std::collections::HashSet;
 
@@ -971,16 +983,47 @@ mod tests {
     }
 
     #[test]
-    #[cfg(feature = "prince")]
     fn invalid_glyph_id() {
         // Test to ensure that invalid glyph ids don't panic when subsetting
-        let buffer = read_fixture("../../../tests/data/fonts/HardGothicNormal.ttf");
+        let buffer = read_fixture("tests/fonts/opentype/Klei.otf");
         let opentype_file = ReadScope::new(&buffer).read::<OpenTypeFont<'_>>().unwrap();
-        let glyph_ids = [0, 9999];
+        let mut glyph_ids = [0, 9999];
 
-        match subset(&opentype_file.table_provider(0).unwrap(), &glyph_ids, None) {
+        match subset(&opentype_file.table_provider(0).unwrap(), &mut glyph_ids) {
             Err(ReadWriteError::Read(ParseError::BadIndex)) => {}
-            _ => panic!("expected ReadWriteError::Read(ParseError::BadIndex) got somthing else"),
+            err => panic!(
+                "expected ReadWriteError::Read(ParseError::BadIndex) got {:?}",
+                err
+            ),
+        }
+    }
+
+    #[test]
+    fn empty_mappings_to_keep() {
+        // Test to ensure that an empty mappings to keep doesn't panic when subsetting
+        let buffer = read_fixture("tests/fonts/opentype/SourceCodePro-Regular.otf");
+        let opentype_file = ReadScope::new(&buffer).read::<OpenTypeFont<'_>>().unwrap();
+        // glyph 118 is not Unicode, so does not end up in the mappings to keep
+        let mut glyph_ids = [0, 118];
+        let subset_font_data =
+            subset(&opentype_file.table_provider(0).unwrap(), &mut glyph_ids).unwrap();
+
+        let opentype_file = ReadScope::new(&subset_font_data)
+            .read::<OpenTypeFont<'_>>()
+            .unwrap();
+        let font = Font::new(opentype_file.table_provider(0).unwrap())
+            .unwrap()
+            .unwrap();
+        let cmap = ReadScope::new(font.cmap_subtable_data())
+            .read::<CmapSubtable<'_>>()
+            .unwrap();
+
+        // If mappings_to_keep is empty a mac roman cmap sub-table is created, which doesn't
+        // care that it's empty.
+        if let CmapSubtable::Format0 { glyph_id_array, .. } = cmap {
+            assert!(glyph_id_array.iter().all(|x| x == 0));
+        } else {
+            panic!("expected cmap sub-table format 0");
         }
     }
 
