@@ -6,12 +6,16 @@
 //!
 //! — <https://docs.microsoft.com/en-us/typography/opentype/spec/cmap>
 
+pub(crate) mod subset;
+
 use std::collections::HashMap;
 use std::convert::TryFrom;
 
 use itertools::izip;
 
-use crate::binary::read::{CheckIndex, ReadArray, ReadBinary, ReadCtxt, ReadFrom, ReadScope};
+use crate::binary::read::{
+    CheckIndex, ReadArray, ReadArrayIter, ReadBinary, ReadCtxt, ReadFrom, ReadScope,
+};
 use crate::binary::write::{WriteBinary, WriteContext};
 use crate::binary::{I16Be, U16Be, U32Be, U8};
 use crate::error::{ParseError, WriteError};
@@ -49,6 +53,11 @@ impl EncodingId {
 
     pub const MACINTOSH_APPLE_ROMAN: EncodingId = EncodingId(0);
     pub const MACINTOSH_UNICODE_UCS4: EncodingId = EncodingId(4);
+
+    /// Unicode 2.0 and onwards semantics, Unicode BMP only
+    pub const UNICODE_BMP: EncodingId = EncodingId(3);
+    /// Unicode 2.0 and onwards semantics, Unicode full repertoire
+    pub const UNICODE_FULL: EncodingId = EncodingId(4);
 }
 
 pub struct Cmap<'a> {
@@ -58,8 +67,8 @@ pub struct Cmap<'a> {
 
 #[derive(Copy, Clone)]
 pub struct EncodingRecord {
-    pub platform_id: u16,
-    pub encoding_id: u16,
+    pub platform_id: PlatformId,
+    pub encoding_id: EncodingId,
     pub offset: u32,
 }
 
@@ -74,14 +83,7 @@ pub enum CmapSubtable<'a> {
         sub_headers: ReadArray<'a, SubHeader>,
         sub_headers_scope: ReadScope<'a>,
     },
-    Format4 {
-        language: u16,
-        end_codes: ReadArray<'a, U16Be>,
-        start_codes: ReadArray<'a, U16Be>,
-        id_deltas: ReadArray<'a, I16Be>,
-        id_range_offsets: ReadArray<'a, U16Be>,
-        glyph_id_array: ReadArray<'a, U16Be>,
-    },
+    Format4(CmapSubtableFormat4<'a>),
     Format6 {
         language: u16,
         first_code: u16,
@@ -107,6 +109,16 @@ pub struct SubHeader {
     id_range_offset: u16,
 }
 
+#[derive(Debug, Clone)]
+pub struct CmapSubtableFormat4<'a> {
+    pub language: u16,
+    pub end_codes: ReadArray<'a, U16Be>,
+    pub start_codes: ReadArray<'a, U16Be>,
+    pub id_deltas: ReadArray<'a, I16Be>,
+    pub id_range_offsets: ReadArray<'a, U16Be>,
+    pub glyph_id_array: ReadArray<'a, U16Be>,
+}
+
 #[derive(Debug, Copy, Clone, PartialEq, PartialOrd, Hash)]
 struct Format4Calculator {
     seg_count: u16,
@@ -114,9 +126,9 @@ struct Format4Calculator {
 
 #[derive(Debug, Copy, Clone, PartialEq, PartialOrd, Hash)]
 pub struct SequentialMapGroup {
-    start_char_code: u32,
-    end_char_code: u32,
-    start_glyph_id: u32,
+    pub(crate) start_char_code: u32,
+    pub(crate) end_char_code: u32,
+    pub(crate) start_glyph_id: u32,
 }
 
 impl<'a> ReadBinary<'a> for Cmap<'a> {
@@ -139,8 +151,8 @@ impl<'a> ReadFrom<'a> for EncodingRecord {
     type ReadType = (U16Be, U16Be, U32Be);
     fn from((platform_id, encoding_id, offset): (u16, u16, u32)) -> Self {
         EncodingRecord {
-            platform_id,
-            encoding_id,
+            platform_id: PlatformId(platform_id),
+            encoding_id: EncodingId(encoding_id),
             offset,
         }
     }
@@ -201,14 +213,14 @@ impl<'a> ReadBinary<'a> for CmapSubtable<'a> {
                 ctxt.check((remaining & 1) == 0)?;
                 let num_indices = remaining >> 1;
                 let glyph_id_array = ctxt.read_array::<U16Be>(num_indices)?;
-                Ok(CmapSubtable::Format4 {
+                Ok(CmapSubtable::Format4(CmapSubtableFormat4 {
                     language,
                     end_codes,
                     start_codes,
                     id_deltas,
                     id_range_offsets,
                     glyph_id_array,
-                })
+                }))
             }
             6 => {
                 let _length = ctxt.read_u16be()?;
@@ -269,14 +281,14 @@ impl<'a> WriteBinary<&Self> for CmapSubtable<'a> {
                 // not be generated for a subset font (the most common path for font writing)
                 return Err(WriteError::NotImplemented);
             }
-            CmapSubtable::Format4 {
+            CmapSubtable::Format4(CmapSubtableFormat4 {
                 language,
                 end_codes,
                 start_codes,
                 id_deltas,
                 id_range_offsets,
                 glyph_id_array,
-            } => {
+            }) => {
                 let start = ctxt.bytes_written();
                 let calc = Format4Calculator {
                     seg_count: u16::try_from(start_codes.len())?,
@@ -345,6 +357,33 @@ impl<'a> WriteBinary<&Self> for CmapSubtable<'a> {
     }
 }
 
+impl<'a, 'b> Format4 for &'a CmapSubtableFormat4<'b> {
+    type U16Iter = ReadArrayIter<'b, U16Be>;
+    type I16Iter = ReadArrayIter<'b, I16Be>;
+
+    fn end_codes(self) -> Self::U16Iter {
+        self.end_codes.iter()
+    }
+
+    fn start_codes(self) -> Self::U16Iter {
+        self.start_codes.iter()
+    }
+
+    fn id_deltas(self) -> Self::I16Iter {
+        self.id_deltas.iter()
+    }
+
+    fn id_range_offsets(self) -> Self::U16Iter {
+        self.id_range_offsets.iter()
+    }
+
+    fn glyph_id_array_get(self, index: usize) -> Result<u16, ParseError> {
+        self.glyph_id_array
+            .check_index(index)
+            .map(|()| self.glyph_id_array.get_item(index))
+    }
+}
+
 impl Format4Calculator {
     fn seg_count_x2(self) -> u16 {
         2 * self.seg_count
@@ -440,7 +479,7 @@ impl<'a> Cmap<'a> {
     pub fn find_subtable_for_platform(&self, platform_id: PlatformId) -> Option<EncodingRecord> {
         self.encoding_records
             .iter()
-            .find(|record| record.platform_id == platform_id.0)
+            .find(|record| record.platform_id == platform_id)
     }
 
     /// Find the first encoding record for the given `platform_id` and `encoding_id`
@@ -449,9 +488,9 @@ impl<'a> Cmap<'a> {
         platform_id: PlatformId,
         encoding_id: EncodingId,
     ) -> Option<EncodingRecord> {
-        self.encoding_records.iter().find(|record| {
-            record.platform_id == platform_id.0 && record.encoding_id == encoding_id.0
-        })
+        self.encoding_records
+            .iter()
+            .find(|record| record.platform_id == platform_id && record.encoding_id == encoding_id)
     }
 
     pub fn encoding_records(&self) -> impl Iterator<Item = EncodingRecord> + 'a {
@@ -483,15 +522,18 @@ impl<'a> CmapSubtable<'a> {
                 let high_byte = ((ch >> 8) & 0xff) as u8;
                 let low_byte = ((ch) & 0xff) as u8;
 
-                let header_index_byte = if high_byte == 0
-                    && sub_header_keys
-                        .check_index(usize::from(low_byte))
-                        .map(|_| sub_header_keys.get_item(usize::from(low_byte)))?
-                        == 0
-                {
-                    usize::from(low_byte)
-                } else {
-                    usize::from(high_byte)
+                let header_index_byte = {
+                    let low_byte = usize::from(low_byte);
+                    if high_byte == 0
+                        && sub_header_keys
+                            .check_index(low_byte)
+                            .map(|_| sub_header_keys.get_item(low_byte))?
+                            == 0
+                    {
+                        low_byte
+                    } else {
+                        usize::from(high_byte)
+                    }
                 };
 
                 let sub_header_key = usize::from(
@@ -523,41 +565,7 @@ impl<'a> CmapSubtable<'a> {
 
                 Ok(Some(glyph_id))
             }
-            CmapSubtable::Format4 {
-                ref end_codes,
-                ref start_codes,
-                ref id_deltas,
-                ref id_range_offsets,
-                ref glyph_id_array,
-                ..
-            } => {
-                for i in 0..end_codes.len() {
-                    // Find segment that contains `ch`
-                    let end_code = u32::from(end_codes.get_item(i));
-                    let start_code = u32::from(start_codes.get_item(i));
-                    if start_code <= ch && ch <= end_code {
-                        // This segment contains ch
-                        let id_delta = i32::from(id_deltas.get_item(i));
-                        let id_range_offset = id_range_offsets.get_item(i);
-                        let glyph_id = if id_range_offset == 0 {
-                            // If the idRangeOffset is 0, the idDelta value is added directly to
-                            // the character code offset (i.e. idDelta[i] + c) to get the
-                            // corresponding glyph index. The idDelta arithmetic is modulo 65536.
-                            (((ch as i32) + id_delta) as u32) & 0xFFFF
-                        } else {
-                            let index = offset_to_index(
-                                i,
-                                id_range_offset,
-                                ch - start_code,
-                                id_range_offsets.len(),
-                            )?;
-                            ((i32::from(glyph_id_array.get_item(index)) + id_delta) as u32) & 0xFFFF
-                        };
-                        return Ok(Some(glyph_id as u16));
-                    }
-                }
-                Ok(None)
-            }
+            CmapSubtable::Format4(ref format4) => format4.map_glyph(ch),
             CmapSubtable::Format6 {
                 first_code,
                 ref glyph_id_array,
@@ -627,21 +635,21 @@ impl<'a> CmapSubtable<'a> {
             // that uses a different format, which would be selected first. As a result support
             // for it is not yet implemented.
             CmapSubtable::Format2 { .. } => None,
-            CmapSubtable::Format4 {
+            CmapSubtable::Format4(CmapSubtableFormat4 {
                 language,
                 end_codes,
                 start_codes,
                 id_deltas,
                 id_range_offsets,
                 glyph_id_array,
-            } => Some(OwnedCmapSubtable::Format4 {
+            }) => Some(OwnedCmapSubtable::Format4(owned::CmapSubtableFormat4 {
                 language: *language,
                 end_codes: end_codes.to_vec(),
                 start_codes: start_codes.to_vec(),
                 id_deltas: id_deltas.to_vec(),
                 id_range_offsets: id_range_offsets.to_vec(),
                 glyph_id_array: glyph_id_array.to_vec(),
-            }),
+            })),
             CmapSubtable::Format6 {
                 language,
                 first_code,
@@ -660,10 +668,12 @@ impl<'a> CmapSubtable<'a> {
                 start_char_code: *start_char_code,
                 glyph_id_array: glyph_id_array.to_vec(),
             }),
-            CmapSubtable::Format12 { language, groups } => Some(OwnedCmapSubtable::Format12 {
-                language: *language,
-                groups: groups.to_vec(),
-            }),
+            CmapSubtable::Format12 { language, groups } => {
+                Some(OwnedCmapSubtable::Format12(owned::CmapSubtableFormat12 {
+                    language: *language,
+                    groups: groups.to_vec(),
+                }))
+            }
         }
     }
 
@@ -676,96 +686,243 @@ impl<'a> CmapSubtable<'a> {
     ///
     /// This method primarily exists to support [GlyphNames](crate::glyph_info::GlyphNames).
     pub(crate) fn mappings(&self) -> Result<HashMap<u16, u32>, ParseError> {
+        let mut mappings = HashMap::with_capacity(self.size_hint());
+        self.mappings_fn(|ch, gid| {
+            mappings.entry(gid).or_insert(ch);
+        })?;
+        Ok(mappings)
+    }
+
+    /// Extract all the mappings from the sub-table.
+    pub fn mappings_fn(&self, mut callback: impl FnMut(u32, u16)) -> Result<(), ParseError> {
         match self {
             CmapSubtable::Format0 {
                 language: _,
                 glyph_id_array,
             } => {
-                let mut mappings = HashMap::with_capacity(glyph_id_array.len());
                 for (ch, gid) in glyph_id_array.iter().enumerate() {
                     // cast is safe as format 0 can only contain 256 glyphs
-                    mappings.entry(u16::from(gid)).or_insert(ch as u32);
+                    callback(ch as u32, u16::from(gid))
                 }
-                Ok(mappings)
             }
-            // It's unlikely that a sub-table using format 2 would be selected for mappings as most
-            // fonts that contain format 2 would probably contain a platform/encoding combination
-            // that uses a different format, which would be selected first. As a result support
-            // for it is not yet implemented.
-            CmapSubtable::Format2 { .. } => return Err(ParseError::NotImplemented),
-            CmapSubtable::Format4 {
-                language: _,
-                end_codes,
-                start_codes,
-                id_deltas,
-                id_range_offsets,
-                glyph_id_array,
+            CmapSubtable::Format2 {
+                sub_header_keys,
+                sub_headers,
+                sub_headers_scope,
+                ..
             } => {
-                let mut mappings = HashMap::with_capacity(glyph_id_array.len());
-                let zipped = izip!(
-                    start_codes.iter(),
-                    end_codes.iter(),
-                    id_deltas.iter(),
-                    id_range_offsets.iter()
-                );
-                for (i, (start_code, end_code, id_delta, id_range_offset)) in zipped.enumerate() {
-                    for (offset_from_start, ch) in (start_code..=end_code).enumerate() {
-                        let glyph_id = if id_range_offset == 0 {
-                            ((i32::from(ch) + i32::from(id_delta)) & 0xFFFF) as u16
-                        } else {
-                            let index = offset_to_index(
-                                i,
-                                id_range_offset,
-                                offset_from_start as u32,
-                                id_range_offsets.len(),
-                            )?;
-                            glyph_id_array.check_index(index)?;
-                            ((i32::from(glyph_id_array.get_item(index)) + i32::from(id_delta))
-                                & 0xFFFF) as u16
-                        };
-                        mappings.entry(glyph_id).or_insert(u32::from(ch));
+                for high_byte in 0u8..=255 {
+                    let sub_header_key = usize::from(
+                        // value is subHeader index × 8.
+                        sub_header_keys
+                            .check_index(usize::from(high_byte))
+                            .map(|_| sub_header_keys.get_item(usize::from(high_byte)))?
+                            / 8,
+                    );
+                    let sub_header = sub_headers
+                        .check_index(sub_header_key)
+                        .map(|_| sub_headers.get_item(sub_header_key))?;
+
+                    // TODO: Reduce duplication
+                    if sub_header_key == 0 {
+                        if !sub_header.contains(u16::from(high_byte)) {
+                            continue; // .notdef
+                        }
+
+                        let glyph_id_index = u16::from(high_byte) - sub_header.first_code;
+
+                        let glyph_index_sub_array =
+                            sub_header.glyph_index_sub_array(sub_header_key, sub_headers_scope)?;
+                        let mut glyph_id = glyph_index_sub_array
+                            .check_index(usize::from(glyph_id_index))
+                            .map(|_| glyph_index_sub_array.get_item(usize::from(glyph_id_index)))?;
+
+                        if glyph_id != 0 {
+                            // The idDelta arithmetic is modulo 65536.
+                            glyph_id =
+                                ((glyph_id as isize + sub_header.id_delta as isize) & 0xffff) as u16
+                        }
+                        callback(u32::from(high_byte), glyph_id);
+                    } else {
+                        for glyph_id_index in 0..sub_header.entry_count {
+                            // FIXME: u8/u16
+                            let low_byte = glyph_id_index + sub_header.first_code;
+
+                            let glyph_index_sub_array = sub_header
+                                .glyph_index_sub_array(sub_header_key, sub_headers_scope)?;
+                            let mut glyph_id = glyph_index_sub_array
+                                .check_index(usize::from(glyph_id_index))
+                                .map(|_| {
+                                    glyph_index_sub_array.get_item(usize::from(glyph_id_index))
+                                })?;
+
+                            if glyph_id != 0 {
+                                // The idDelta arithmetic is modulo 65536.
+                                glyph_id = ((glyph_id as isize + sub_header.id_delta as isize)
+                                    & 0xffff) as u16
+                            }
+                            callback((u32::from(high_byte) << 8) | u32::from(low_byte), glyph_id);
+                        }
                     }
                 }
-                Ok(mappings)
             }
+            CmapSubtable::Format4(format4) => format4.mappings_fn(callback)?,
             CmapSubtable::Format6 {
                 language: _,
                 first_code,
                 glyph_id_array,
             } => {
-                let mut mappings = HashMap::with_capacity(glyph_id_array.len());
                 for (index, gid) in glyph_id_array.iter().enumerate() {
                     // cast is safe as the entryCount of the glyphIdArray is a 16-bit value
-                    mappings
-                        .entry(gid)
-                        .or_insert(u32::from(*first_code) + index as u32);
+                    callback(u32::from(*first_code) + index as u32, gid)
                 }
-                Ok(mappings)
             }
             CmapSubtable::Format10 {
                 language: _,
                 start_char_code,
                 glyph_id_array,
             } => {
-                let mut mappings = HashMap::with_capacity(glyph_id_array.len());
                 for (index, gid) in glyph_id_array.iter().enumerate() {
                     let index = u32::try_from(index)?;
-                    mappings.entry(gid).or_insert(*start_char_code + index);
+                    callback(*start_char_code + index, gid)
                 }
-                Ok(mappings)
             }
             CmapSubtable::Format12 { groups, .. } => {
-                let mut mappings = HashMap::new();
                 for record in groups.iter() {
                     for (i, ch) in (record.start_char_code..=record.end_char_code).enumerate() {
-                        mappings
-                            .entry(u16::try_from(record.start_glyph_id)? + u16::try_from(i)?)
-                            .or_insert(ch);
+                        callback(
+                            ch,
+                            u16::try_from(record.start_glyph_id)? + u16::try_from(i)?,
+                        )
                     }
                 }
-                Ok(mappings)
             }
         }
+
+        Ok(())
+    }
+
+    /// A hint as to the number of mappings contained in this sub-table.
+    ///
+    /// For some formats it will be the exact size, for others it will be underestimated.
+    pub(crate) fn size_hint(&self) -> usize {
+        match self {
+            CmapSubtable::Format0 { glyph_id_array, .. } => glyph_id_array.len(),
+            CmapSubtable::Format2 { .. } => 0, // TODO: Implement if needed in mappings_fn
+            CmapSubtable::Format4(CmapSubtableFormat4 { glyph_id_array, .. }) => {
+                glyph_id_array.len()
+            }
+            CmapSubtable::Format6 { glyph_id_array, .. } => glyph_id_array.len(),
+            CmapSubtable::Format10 { glyph_id_array, .. } => glyph_id_array.len(),
+            CmapSubtable::Format12 { groups, .. } => groups
+                .iter()
+                .map(|group| {
+                    let start_char_code = group.start_char_code as usize;
+                    let end_char_code = group.end_char_code as usize;
+                    end_char_code.saturating_sub(start_char_code)
+                })
+                .sum(),
+        }
+    }
+}
+
+trait Format4 {
+    type U16Iter: ExactSizeIterator<Item = u16>;
+    type I16Iter: ExactSizeIterator<Item = i16>;
+
+    fn end_codes(self) -> Self::U16Iter;
+    fn start_codes(self) -> Self::U16Iter;
+    fn id_deltas(self) -> Self::I16Iter;
+    fn id_range_offsets(self) -> Self::U16Iter;
+    fn glyph_id_array_get(self, index: usize) -> Result<u16, ParseError>;
+
+    fn map_glyph(self, ch: u32) -> Result<Option<u16>, ParseError>
+    where
+        Self: Sized + Copy,
+    {
+        // Format 4 sub-tables can only map a 16-bit character range
+        let ch = u16::try_from(ch)?;
+        let zipped = izip!(
+            self.start_codes(),
+            self.end_codes(),
+            self.id_deltas(),
+            self.id_range_offsets()
+        );
+        for (i, (start_code, end_code, id_delta, id_range_offset)) in zipped.enumerate() {
+            // Find segment that contains `ch`
+            if start_code <= ch && ch <= end_code {
+                // This segment contains ch
+                let glyph_id = self.glyph_id_for_id_range_offset(
+                    id_range_offset,
+                    ch,
+                    id_delta,
+                    i,
+                    ch - start_code,
+                )?;
+                return Ok(Some(glyph_id));
+            }
+        }
+        Ok(None)
+    }
+
+    fn mappings_fn(self, mut callback: impl FnMut(u32, u16)) -> Result<(), ParseError>
+    where
+        Self: Sized + Copy,
+    {
+        let zipped = izip!(
+            self.start_codes(),
+            self.end_codes(),
+            self.id_deltas(),
+            self.id_range_offsets()
+        );
+        for (i, (start_code, end_code, id_delta, id_range_offset)) in zipped.enumerate() {
+            for (offset_from_start, ch) in (start_code..=end_code).enumerate() {
+                let glyph_id = self.glyph_id_for_id_range_offset(
+                    id_range_offset,
+                    ch,
+                    id_delta,
+                    i,
+                    offset_from_start as u16,
+                )?;
+                callback(u32::from(ch), glyph_id)
+            }
+        }
+
+        Ok(())
+    }
+
+    fn glyph_id_for_id_range_offset(
+        self,
+        mut id_range_offset: u16,
+        ch: u16,
+        id_delta: i16,
+        segment_index: usize,
+        start_code_offset: u16,
+    ) -> Result<u16, ParseError>
+    where
+        Self: Sized + Copy,
+    {
+        // Work around Fontographer bug
+        // https://github.com/adobe-type-tools/afdko/blob/01a35dacc9e8d1735b7f752f3232d38c34e6f843/c/shared/source/ttread/ttread.c#L1885
+        if id_range_offset == 0xFFFF {
+            id_range_offset = 0;
+        }
+
+        if id_range_offset == 0 {
+            // If the idRangeOffset is 0, the idDelta value is added directly to
+            // the character code offset (i.e. idDelta[i] + c) to get the
+            // corresponding glyph index. The idDelta arithmetic is modulo 65536.
+            Ok(((i32::from(ch) + i32::from(id_delta)) & 0xFFFF) as u16)
+        } else {
+            let index = offset_to_index(
+                segment_index,
+                id_range_offset,
+                start_code_offset,
+                self.id_range_offsets().len(),
+            )?;
+            let glyph_id = self.glyph_id_array_get(index)?;
+            Ok(((i32::from(glyph_id) + i32::from(id_delta)) & 0xFFFF) as u16)
+        }
     }
 }
 
@@ -773,14 +930,14 @@ impl<'a> CmapSubtable<'a> {
 fn offset_to_index(
     i: usize,
     id_range_offset: u16,
-    start_code_offset: u32,
+    start_code_offset: u16,
     id_range_offsets_len: usize,
 ) -> Result<usize, ParseError> {
     // Offset into `id_range_offsets` that `i` represents, * 2 for 16-bit values in the array.
     // cast is safe as i is segment index, which is a u16
     let offset_in_id_range_offsets = u32::from(id_range_offset) + i as u32 * 2;
     // Offset into `glyph_id_array` is offset from `start_code` of `ch` * 2 for 16-bit glyph ids.
-    let glyph_id_offset = offset_in_id_range_offsets + start_code_offset * 2;
+    let glyph_id_offset = offset_in_id_range_offsets + u32::from(start_code_offset) * 2;
     // Bounds check, cast is safe as id_range_offsets has max segCount items, a 16-bit value.
     if glyph_id_offset >= id_range_offsets_len as u32 * 2 && (glyph_id_offset & 1) == 0 {
         // Turn the offsets into an index
@@ -792,8 +949,8 @@ fn offset_to_index(
 
 pub mod owned {
     use super::{
-        offset_to_index, size, Format4Calculator, I16Be, ParseError, SequentialMapGroup, TryFrom,
-        U16Be, U32Be, WriteBinary, WriteContext, WriteError,
+        size, EncodingId, Format4, Format4Calculator, I16Be, ParseError, PlatformId,
+        SequentialMapGroup, TryFrom, U16Be, U32Be, WriteBinary, WriteContext, WriteError,
     };
 
     #[derive(Debug, Clone, PartialEq)]
@@ -803,8 +960,8 @@ pub mod owned {
 
     #[derive(Debug, Clone, PartialEq)]
     pub struct EncodingRecord {
-        pub platform_id: u16,
-        pub encoding_id: u16,
+        pub platform_id: PlatformId,
+        pub encoding_id: EncodingId,
         pub sub_table: CmapSubtable,
     }
 
@@ -814,14 +971,7 @@ pub mod owned {
             language: u16,
             glyph_id_array: Box<[u8; 256]>,
         },
-        Format4 {
-            language: u16,
-            end_codes: Vec<u16>,
-            start_codes: Vec<u16>,
-            id_deltas: Vec<i16>,
-            id_range_offsets: Vec<u16>,
-            glyph_id_array: Vec<u16>,
-        },
+        Format4(CmapSubtableFormat4),
         Format6 {
             language: u16,
             first_code: u16,
@@ -832,10 +982,23 @@ pub mod owned {
             start_char_code: u32,
             glyph_id_array: Vec<u16>,
         },
-        Format12 {
-            language: u32,
-            groups: Vec<SequentialMapGroup>,
-        },
+        Format12(CmapSubtableFormat12),
+    }
+
+    #[derive(Debug, Clone, PartialEq)]
+    pub struct CmapSubtableFormat4 {
+        pub language: u16,
+        pub end_codes: Vec<u16>,
+        pub start_codes: Vec<u16>,
+        pub id_deltas: Vec<i16>,
+        pub id_range_offsets: Vec<u16>,
+        pub glyph_id_array: Vec<u16>,
+    }
+
+    #[derive(Debug, Clone, PartialEq)]
+    pub struct CmapSubtableFormat12 {
+        pub language: u32,
+        pub groups: Vec<SequentialMapGroup>,
     }
 
     impl CmapSubtable {
@@ -853,41 +1016,7 @@ pub mod owned {
                         Ok(None)
                     }
                 }
-                CmapSubtable::Format4 {
-                    ref end_codes,
-                    ref start_codes,
-                    ref id_deltas,
-                    ref id_range_offsets,
-                    ref glyph_id_array,
-                    ..
-                } => {
-                    for i in 0..end_codes.len() {
-                        // Find segment that contains `ch`
-                        let end_code = u32::from(end_codes[i]);
-                        let start_code = u32::from(start_codes[i]);
-                        if start_code <= ch && ch <= end_code {
-                            // This segment contains ch
-                            let id_delta = i32::from(id_deltas[i]);
-                            let id_range_offset = id_range_offsets[i];
-                            let glyph_id = if id_range_offset == 0 {
-                                // If the idRangeOffset is 0, the idDelta value is added directly to
-                                // the character code offset (i.e. idDelta[i] + c) to get the
-                                // corresponding glyph index. The idDelta arithmetic is modulo 65536.
-                                (((ch as i32) + id_delta) as u32) & 0xFFFF
-                            } else {
-                                let index = offset_to_index(
-                                    i,
-                                    id_range_offset,
-                                    ch - start_code,
-                                    id_range_offsets.len(),
-                                )?;
-                                ((i32::from(glyph_id_array[index]) + id_delta) as u32) & 0xFFFF
-                            };
-                            return Ok(Some(glyph_id as u16));
-                        }
-                    }
-                    Ok(None)
-                }
+                CmapSubtable::Format4(ref format4) => format4.map_glyph(ch),
                 CmapSubtable::Format6 {
                     first_code,
                     ref glyph_id_array,
@@ -923,7 +1052,7 @@ pub mod owned {
                         Ok(None)
                     }
                 }
-                CmapSubtable::Format12 { ref groups, .. } => {
+                CmapSubtable::Format12(CmapSubtableFormat12 { ref groups, .. }) => {
                     for group in groups {
                         if group.start_char_code <= ch && ch <= group.end_char_code {
                             let glyph_id = group.start_glyph_id + (ch - group.start_char_code);
@@ -936,6 +1065,34 @@ pub mod owned {
         }
     }
 
+    impl<'a> Format4 for &'a CmapSubtableFormat4 {
+        type U16Iter = std::iter::Copied<std::slice::Iter<'a, u16>>;
+        type I16Iter = std::iter::Copied<std::slice::Iter<'a, i16>>;
+
+        fn end_codes(self) -> Self::U16Iter {
+            self.end_codes.iter().copied()
+        }
+
+        fn start_codes(self) -> Self::U16Iter {
+            self.start_codes.iter().copied()
+        }
+
+        fn id_deltas(self) -> Self::I16Iter {
+            self.id_deltas.iter().copied()
+        }
+
+        fn id_range_offsets(self) -> Self::U16Iter {
+            self.id_range_offsets.iter().copied()
+        }
+
+        fn glyph_id_array_get(self, index: usize) -> Result<u16, ParseError> {
+            self.glyph_id_array
+                .get(index)
+                .copied()
+                .ok_or(ParseError::BadIndex)
+        }
+    }
+
     impl<'a> WriteBinary<Self> for Cmap {
         type Output = ();
 
@@ -947,8 +1104,8 @@ pub mod owned {
             // encoding records
             let mut offsets = Vec::with_capacity(table.encoding_records.len());
             for record in &table.encoding_records {
-                U16Be::write(ctxt, record.platform_id)?;
-                U16Be::write(ctxt, record.encoding_id)?;
+                U16Be::write(ctxt, record.platform_id.0)?;
+                U16Be::write(ctxt, record.encoding_id.0)?;
                 let offset = ctxt.placeholder::<U32Be, _>()?;
                 offsets.push(offset);
             }
@@ -979,14 +1136,14 @@ pub mod owned {
                     U16Be::write(ctxt, language)?;
                     ctxt.write_bytes(glyph_id_array.as_ref())?;
                 }
-                CmapSubtable::Format4 {
+                CmapSubtable::Format4(CmapSubtableFormat4 {
                     language,
                     end_codes,
                     start_codes,
                     id_deltas,
                     id_range_offsets,
                     glyph_id_array,
-                } => {
+                }) => {
                     let start = ctxt.bytes_written();
                     let calc = Format4Calculator {
                         seg_count: u16::try_from(start_codes.len())?,
@@ -1038,7 +1195,7 @@ pub mod owned {
                     ctxt.write_vec::<U16Be>(glyph_id_array)?;
                     ctxt.write_placeholder(length, u32::try_from(ctxt.bytes_written() - start)?)?;
                 }
-                CmapSubtable::Format12 { language, groups } => {
+                CmapSubtable::Format12(CmapSubtableFormat12 { language, groups }) => {
                     let start = ctxt.bytes_written();
 
                     U16Be::write(ctxt, 12u16)?; // format
@@ -1145,6 +1302,54 @@ mod tests {
         );
     }
 
+    #[test]
+    #[cfg(feature = "prince")]
+    fn test_mappings_format2() {
+        with_cmap_subtable(
+            "../../../tests/data/fonts/HardGothicNormal.ttf",
+            PlatformId::WINDOWS,
+            EncodingId::WINDOWS_BIG5,
+            |cmap_subtable| {
+                if !matches!(cmap_subtable, CmapSubtable::Format2 { .. }) {
+                    panic!("expected CmapSubtable::Format2");
+                }
+
+                let mappings = cmap_subtable.mappings().unwrap();
+                // This would be a good place for non-ASCII idents when we're on 1.53 or newer
+                // let 世 = cmap_subtable.map_glyph(12).unwrap().unwrap(); // 世 U+4E16
+                // let 丈 = cmap_subtable.map_glyph(6).unwrap().unwrap(); // 丈 U+4E08
+                let a = cmap_subtable.map_glyph(13).unwrap().unwrap(); // 丕 U+4E15
+                let b = cmap_subtable.map_glyph(44).unwrap().unwrap(); // 乾 U+4E7E
+                assert_eq!(mappings[&a], 13);
+                assert_eq!(mappings[&b], 44);
+            },
+        );
+    }
+
+    #[test]
+    #[cfg(feature = "prince")]
+    fn test_mappings_format2_chao_yan_ze_cu_hei_tif() {
+        with_cmap_subtable(
+            "../../../tests/data/fonts/big5/ChaoYanZeCuHeiTif-1.ttf",
+            PlatformId::WINDOWS,
+            EncodingId::WINDOWS_BIG5,
+            |cmap_subtable| {
+                if !matches!(cmap_subtable, CmapSubtable::Format2 { .. }) {
+                    panic!("expected CmapSubtable::Format2");
+                }
+
+                // This is checking that one and two-byte characters work as well as certain
+                // entries that should not be present are absent (E.g. 0x220). This test was
+                // added after `mappings_fn` for cmap format 2 was written. Prior to the change
+                // many entries (such as 0x220) were included that should not have been.
+                let mappings = cmap_subtable.mappings().unwrap();
+                assert_eq!(mappings[&85], 0x54);
+                assert_eq!(mappings[&461], 0xA26F);
+                assert!(mappings.values().find(|&&ch| ch == 0x220).is_none());
+            },
+        );
+    }
+
     #[test]
     fn test_mappings_format4() {
         with_cmap_subtable(
